import torch 

'''
a naive implementation of hotrg 
'''

def gauge(M, Dcut, side):
    if side=='l':
        D2 = M.shape[0]
        M = M.view(D2, -1)
    elif side=='r':
        D2 = M.shape[3]
        M = M.permute(3,0,1,2).view(D2, -1)
    M2 = M@M.t()

    w, v = torch.symeig(M2, eigenvectors=True)
    Dnew = min(D2, Dcut)
    truncation = 1.-w[-Dnew:].sum() / w.sum()
    return v[:, -Dnew:], truncation

def hotrg(T, D, Dcut, no_iter):
    lnZ = 0.0
    truncation_error = 0.0
    for n in range(no_iter):
        f = torch.norm(T)
        lnZ += 2**(no_iter-n)*torch.log(f)
        T = T / f

        #contract up-down
        Dh, Dv = T.shape[0], T.shape[1] # horizontal and vertical bond dimension of the tensor
        M = torch.einsum('ludr,adcb->laucrb', (T, T)).contiguous().view(Dh**2, Dv, Dv, Dh**2)
        vl, el = gauge(M, Dcut,'l')
        vr, er = gauge(M, Dcut,'r')
        U = vl if (el < er) else vr #(Dh**2, Dnew)
        Dnew = U.shape[1]
        truncation_error += min(el, er)
        T = torch.einsum('li,ludr,rj->ujid', (U, M, U)) 

    trace = 0.0 
    for x in range(T.shape[0]):
        for y in range(T.shape[1]):
            trace += T[x, y, y, x]
    lnZ += torch.log(trace)
    return lnZ/2**no_iter, truncation_error

if __name__=='__main__':
    K = torch.tensor([0.44])
    Dcut = 20
    n = 20

    #Boltzmann factor on a bond M=LR^T
    M = torch.stack([torch.cat([torch.exp(K),  torch.exp(-K)]), 
                     torch.cat([torch.exp(-K), torch.exp(K)])
                    ])
    U, S, V = torch.svd(M)
    L = U*torch.sqrt(S) 
    R = V*torch.sqrt(S)

    #            L 
    #            |
    # T =  R^{T}-o-L
    #            |
    #            R^{T}

    T = torch.einsum('ai,aj,ak,al->ijkl', (L, L, R, R)) 
    lnZ, _ = hotrg(T, 2, Dcut, n)
    print (lnZ.item())
