'''
'''
import torch 
torch.set_num_threads(4)
torch.manual_seed(42)

from hotrg2 import hotrg as contraction
#from trg import levin_nave_trg as contraction

if __name__=='__main__':
    import time
    import argparse
    parser = argparse.ArgumentParser(description='')
    parser.add_argument("-D", type=int, default=2, help="D")
    parser.add_argument("-Dcut", type=int, default=20, help="Dcut")
    parser.add_argument("-beta", type=float, default=0.22, help="beta")
    parser.add_argument("-Niter", type=int, default=20, help="Niter")

    parser.add_argument("-float32", action='store_true', help="use float32")
    parser.add_argument("-cuda", type=int, default=-1, help="use GPU")
    args = parser.parse_args()
    device = torch.device("cpu" if args.cuda<0 else "cuda:"+str(args.cuda))
    dtype = torch.float32 if args.float32 else torch.float64

    d = 2 # fixed

    D = args.D
    Dcut = args.Dcut
    Niter = args.Niter
    beta = torch.tensor([args.beta], dtype=dtype, device=device).requires_grad_()

    A = torch.nn.Parameter(0.01* torch.randn(d, D**4, dtype=dtype, device=device, requires_grad=True))

    #3D Ising
    c = torch.sqrt(torch.cosh(beta))
    s = torch.sqrt(torch.sinh(beta))
    M = torch.stack([torch.cat([c, s]), torch.cat([c, -s])])
    T = torch.einsum('ai,aj,ak,al,am,an->ijklmn', (M, M, M, M, M, M))
    T = T.view(d, d**4, d)

    optimizer = torch.optim.LBFGS([A], max_iter=50)
    
    def closure():
        optimizer.zero_grad()

        T1 = torch.einsum('xa,xby,yc' , (A,T,A)).view(D,D,D,D, d,d,d,d, D,D,D,D).permute(0,4,8, 1,5,9, 2,6,10, 3,7,11).contiguous().view(D**2*d, D**2*d, D**2*d, D**2*d)
        #double layer 
        T2 = (A.t()@A).view(D, D, D, D, D, D, D, D).permute(0,4, 1,5, 2,6, 3,7).contiguous().view(D**2, D**2, D**2, D**2)

        t0=time.time()
        lnT1, error1 = contraction(T1, D**2*d, Dcut, Niter)
        lnT2, error2 = contraction(T2, D**2, Dcut, Niter)
        loss = (-lnT1 + lnT2)/2**Niter  # loss = -lnZ of Ising
        print ('contraction done {:.3f}s'.format(time.time()-t0))
        print ('truncation error', error1.item(), error2.item())

        t0=time.time()
        loss.backward(retain_graph=True)
        print ('backward done {:.3f}s'.format(time.time()-t0))
        print ('free energy, energy', loss.item(), beta.grad.item()) # En = -d lnZ / d beta
        beta.grad.zero_()
        return loss

    optimizer.step(closure)
